// Filename: depth_volumedata_storage_fs.glsl
// Fragment shader that save depth of fragment into 2D Texture

//layout(pixel_center_integer) in vec4 gl_FragCoord; //to (0.0, 0.0)

uniform sampler3D IMAGE3D;
uniform sampler3D CONTROLVOLUME;
uniform sampler1D TRANSFERFUNCTION1D;
uniform vec4 scaleFactor;
uniform vec3 extents;
uniform float stepSize;
uniform vec3 cursorPosition;
uniform bvec3 drawPlaneSide1;
uniform bvec3 drawPlaneSide2;
uniform bvec3 clippingPlaneSide1;
uniform bvec3 clippingPlaneSide2;
uniform float depth_slicing;
uniform float noise_threshold;
uniform bool flipPlaneSide1;

//global variable
float z_value;

// Find the ray intersection interval with respect 
// to the proxy geometry  (Liang Barsky)
bool intersectBox(in vec4 r_o, in vec4 r_d, in vec4 boxmin, in vec4 boxmax, 
inout float tnear, inout float tfar)
{
    // compute intersection of ray with all six bbox planes (Liang-Barsky)
    vec4 invR = vec4(1.0,1.0,1.0,1.0)/r_d;
    vec4 tbot = invR * (boxmin - r_o);
    vec4 ttop = invR * (boxmax - r_o);

    // re-order intersections to find smallest and largest on each axis
    vec4 tmin = min(ttop, tbot);
    vec4 tmax = max(ttop, tbot);

    if (all(notEqual (tmin.xyz,tmax.xyz))) {
        // find the largest tmin and the smallest tmax
    	float largest_tmin = max(max(tmin.x, tmin.y), max(tmin.x, tmin.z));
    	float smallest_tmax = min(min(tmax.x, tmax.y), min(tmax.x, tmax.z));

    	tnear = largest_tmin;
    	tfar = smallest_tmax;

    	if (tfar >= tnear) return true;
    	else return false;
    } else {
      // r_d is parallel to at least two planes. Just consider the range [0,1].
      // It should be improved.
      tnear = 0.;
      tfar = 1.0;
      return true;
    }
}

// Compute the depht (z value) of the first no transparente fragment 
// returning into normalize screen coordinate space  
void processZvalue(in vec4 geomPos, inout bool data_found){
	vec4 centralizedGeomPos, boundary;
	mat4 mvp = gl_ModelViewProjectionMatrix;
	
	boundary.xyz = extents.xyz;
	boundary.w = 1.0;

	centralizedGeomPos = boundary * geomPos; // geomPos range: [0,1]; centralizedGeomPos range: [-1,1]
	z_value = (mvp * centralizedGeomPos).z;// transform the centralized geomPos
	z_value = 0.5 * z_value + 0.5;//scale in the range [0,1]
	
	data_found = true;
}

//Process all control tags (markups and depths, at this moment)
void processControlTags(in float image3Dscalar, in float TFscalar, 
in vec4 geomPos, inout bool data_found, inout vec4 dstColor){
	int markup;
	float depth;
	vec4 controlTag;

	controlTag = texture3D(CONTROLVOLUME, geomPos.xyz);
	markup = int(controlTag.x * float(INT_FORMAT_BITS));
	
	//all markups (except transparency) are considered valid voxels
	//to be included in depth buffer
	switch (markup){
	case TAG001://transparent
		break;
	case TAG002://painted
		dstColor = vec4(0.0, 1.0, 1.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG003://point of reference
		dstColor = vec4(1.0, 0.0, 1.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG004://point of interest
		dstColor = vec4(1.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG005://line
		dstColor = vec4(0.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG006://eraser
		dstColor = vec4(1.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;		
	default: //if there wasn't a markup tag, begins depth processing 
		depth = controlTag.y;//get voxel's depth //Revised by Ting
		
		//zvalue is set once, only to the first valid voxel in ray composition.
		//so, if some valid voxel was already found, depth buffer from previous voxel
		//won't be ovewritten, ignoring this step
		if(!data_found)
		
			//1) all untagged voxels have depth=0 and must be shown.
			//2) also, voxels deeper (with greater depth) than the slicing mesh have to be shown.
			if(depth == 0. || depth_slicing <= depth)
			
				//a scalar greater than noise threshold determines a valid voxel
				if(image3Dscalar > noise_threshold)
				
					//TFscalar is what the user is seeing indeed,
					//so consider only valid entries from transfer function
					if(TFscalar > 0.0)
						processZvalue(geomPos, data_found);
		break;
	}
}

//Clipping volume with axis aligned planes
void processAxisPlanes(in bool flipPlane, in vec4 geomPos, in vec4 axialColor, 
in vec4 sagittalColor, in vec4 coronalColor, inout bool data_found, 
inout vec4 dstColor){

	//Drawing planes from side 1
	if (drawPlaneSide1[VMTK_AXIAL] == true && geomPos.y < cursorPosition.z + stepSize 
			&& geomPos.y > cursorPosition.z ) {
		dstColor += axialColor;
		processZvalue(geomPos, data_found);
	}
	if (drawPlaneSide1[VMTK_SAGITTAL] == true && geomPos.z < cursorPosition.x + stepSize
			&& geomPos.z > cursorPosition.x ) {
		dstColor += sagittalColor;
		processZvalue(geomPos, data_found);
	}
	if (drawPlaneSide1[VMTK_CORONAL] == true && geomPos.x < cursorPosition.y + stepSize
			&& geomPos.x > cursorPosition.y ) {
		dstColor += coronalColor;
		processZvalue(geomPos, data_found);
	}

	//Clipping planes from side 1
	if (!flipPlane){
		if ((clippingPlaneSide1[VMTK_AXIAL] == true && geomPos.y > cursorPosition.z  + stepSize )
			|| (clippingPlaneSide1[VMTK_SAGITTAL] == true && geomPos.z > cursorPosition.x + stepSize )
			|| (clippingPlaneSide1[VMTK_CORONAL] == true && geomPos.x > cursorPosition.y + stepSize )){

			//no color until ray is passing through clipped portion of volume
			dstColor = vec4(0.0);
			z_value = 1.0;

			//the clipped portion of volume isn't valid data to be 
			//included in ray composition, so no data found until this position
			data_found = false;
		}
	}else {
		if ((clippingPlaneSide1[VMTK_AXIAL] == true && geomPos.y < cursorPosition.z  + stepSize )
			|| (clippingPlaneSide1[VMTK_SAGITTAL] == true && geomPos.z < cursorPosition.x + stepSize )
			|| (clippingPlaneSide1[VMTK_CORONAL] == true && geomPos.x < cursorPosition.y + stepSize )){

			//no color until ray is passing through clipped portion of volume
			dstColor = vec4(0.0);
			z_value = 1.0;

			//the clipped portion of volume isn't valid data to be 
			//included in ray composition, so no data found until this position
			data_found = false;
		}
	}//EndIf FlipPlane
}

void main(void) {

	bool data_found, hit;
	float scalar, t, tnear, tfar;	 
	
	vec4 camDir, dstColor , srcColor;	
	vec4 axialColor, sagittalColor, coronalColor,volExtentMin, volExtentMax;
	vec4 geomDir; //direction vector
	vec4 geomPos, initGeomPos ; //point position
	mat4 mvi ;	
	
	//variable initialization
	mvi = gl_ModelViewMatrixInverse; //matrix model-view-inverse	
	volExtentMin 	= vec4(0.0, 0.0, 0.0, 0.0); //mim value of boundingbox
	volExtentMax 	= vec4(1.0, 1.0, 1.0, 0.0); //max value of boundingbox	
	camDir 			= vec4(0.0, 0.0, -1.0, 0.0); //camera direction for paralell projection
	geomPos 		= vec4(0.0);
	
	dstColor 		= srcColor = vec4(0.0);//initialize color	
	axialColor 		= vec4(0.0, 0.08, 0.08, 0.2); //color of axial plane
	sagittalColor 	= vec4(0.08, 0.0, 0.08, 0.2); //color of sagittal plane
	coronalColor 	= vec4(0.08, 0.08, 0.0, 0.2); //color of coronal plane
	
	z_value = 1.0; // Default depth value	
	data_found = false; //flag used when non transparent voxels are found
	
	//RAY CASTING ALGORITHM
	//(1)determine volume entry position
	initGeomPos = gl_TexCoord[1];
	initGeomPos.w = 1.0;
	
	//(2)compute ray direction
	geomDir = normalize(mvi * camDir) * scaleFactor;
	geomDir.w = 0.0;
	
	//(2.1) foreground and background plane
	hit = intersectBox(initGeomPos , geomDir, volExtentMin, volExtentMax,
			 tnear, tfar);	
	if (!hit) {
        gl_FragColor = vec4(1.0,1.0,1.0,1.0);
       	//gl_FragDepth = z_value;
       return;
    }
			 
	if (tnear < 0.0) 
		tnear = 0.0;  // clamp to near plane	
	t = tnear;
	
	//(3) ray traversal loop
	while (t < tfar && dstColor.a < 0.99) {
		
		//(3.1) Determine entry position
		geomPos = initGeomPos + geomDir * t;
				
		//(3.2) Apply transfer function
		scalar = texture3D(IMAGE3D, geomPos.xyz).r; //recover index Red					
		srcColor  = texture1D(TRANSFERFUNCTION1D, scalar); //Apply transfer function
		srcColor.rgb *= srcColor.a; //Compute associated color of voxel
										
		// Determines depth values of markup fragments		
		processControlTags(scalar, srcColor.a, geomPos, data_found, dstColor);
		
		//draw or clip the volume
		processAxisPlanes(flipPlaneSide1, geomPos, axialColor, sagittalColor,
				 coronalColor, data_found, dstColor);
				 			
		//(3.3) front-to-back composition
		if(data_found){								
			dstColor.rgb += (1.0 - dstColor.a) * srcColor.rgb;
			dstColor.a += (1.0 - dstColor.a) * srcColor.a;
			break; //force to save only first non-transparent voxel
		}
				
		//(3.4) advance ray position along ray direction										
		t += stepSize;//*0.4;			
		
	}//End while
		
	gl_FragColor = vec4(geomPos.x, geomPos.y, geomPos.z, z_value);
	
}