// # Filename: fragmentLens3D.glsl

#extension GL_ARB_texture_rectangle : enable //rectangle texture

uniform sampler3D IMAGE3D;
uniform sampler3D CONTROLVOLUME;
uniform sampler1D TRANSFERFUNCTION1D;
uniform sampler1D TRANSFERFUNCTIONCOLOR1D;
uniform sampler2DRect LENSDEPTHGEOMETRY;
uniform vec4 scaleFactor;
uniform vec3 extents;
uniform vec3 cursorPosition;
uniform bvec3 drawPlaneSide1;
uniform bvec3 drawPlaneSide2;
uniform bvec3 clippingPlaneSide1;
uniform bvec3 clippingPlaneSide2;
uniform float stepSize;
uniform float depth_slicing;
uniform float noise_threshold;

uniform bool useSelTF;
uniform bool lensComposition;
uniform vec3 focalPoint;
uniform float lensZooming;
uniform bool flipPlaneSide1;

float z_value;

// Find the ray intersection interval with respect 
// to the proxy geometry  (Liang Barsky)
bool intersectBox(in vec4 r_o, in vec4 r_d, in vec4 boxmin, in vec4 boxmax, 
inout float tnear, inout float tfar)
{
    // compute intersection of ray with all six bbox planes (Liang-Barsky)
    vec4 invR = vec4(1.0,1.0,1.0,1.0)/r_d;
    vec4 tbot = invR * (boxmin - r_o);
    vec4 ttop = invR * (boxmax - r_o);

    // re-order intersections to find smallest and largest on each axis
    vec4 tmin = min(ttop, tbot);
    vec4 tmax = max(ttop, tbot);

    if (all(notEqual (tmin.xyz,tmax.xyz))) {
        // find the largest tmin and the smallest tmax
    	float largest_tmin = max(max(tmin.x, tmin.y), max(tmin.x, tmin.z));
    	float smallest_tmax = min(min(tmax.x, tmax.y), min(tmax.x, tmax.z));

    	tnear = largest_tmin;
    	tfar = smallest_tmax;

    	if (tfar >= tnear) return true;
    	else return false;
    } else {
      // r_d is parallel to at least two planes. Just consider the range [0,1].
      // It should be improved.
      tnear = 0.;
      tfar = 1.0;
      return true;
    }
}


//Get screen-coordinate depht [0,1] from single world-coordinate point
float getDepthOfSample(in vec4 geomPos){
	vec4 centralizedGeomPos, boundary;
	mat4 mvp;
	float depth;
	
	mvp = gl_ModelViewProjectionMatrix;
	boundary.xyz = extents.xyz;
	boundary.w = 1.0;

	centralizedGeomPos = boundary * geomPos; // geomPos texture range: [0,1] 
	depth = (mvp * centralizedGeomPos).z;// transform the NDC [-1,1]
	depth = 0.5 * depth + 0.5;//put it in range [0,1] DC		
	return depth;
}

//Process geometric position from any voxel,
//copying its depth to depth buffer.
//thus any voxels between a observer and this voxel
//will be ignored from visualization
void processZvalue(in vec4 geomPos, inout bool data_found){
	vec4 centralizedGeomPos, boundary;
	mat4 mvp;
	
	mvp = gl_ModelViewProjectionMatrix;
	boundary.xyz = extents.xyz;
	boundary.w = 1.0;

	centralizedGeomPos = boundary * geomPos; // geomPos range: [0,1]; centralizedGeomPos range: [-1,1]
	z_value = (mvp * centralizedGeomPos).z;// transform the centralized geomPos
	z_value = 0.5 * z_value + 0.5;//put it in range [0,1]
	
	data_found = true;
}

//Process all control tags (markups and depths, at this moment)
void processControlTags(in float image3Dscalar, in float TFscalar, 
in vec4 geomPos, inout bool data_found, inout vec4 dstColor){
	int markup;
	float depth;
	vec4 controlTag;

	controlTag = texture3D(CONTROLVOLUME, geomPos.xyz);
	markup = int(controlTag.r * float(INT_FORMAT_BITS));
	
	//all markups (except transparency) are considered valid voxels
	//to be included in depth buffer
	switch (markup){
	case TAG001://transparent
		break;
	case TAG002://painted
		dstColor = vec4(0.0, 1.0, 1.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG003://point of reference
		dstColor = vec4(1.0, 0.0, 1.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG004://point of interest
		dstColor = vec4(1.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG005://line
		dstColor = vec4(0.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG006://eraser
		dstColor = vec4(1.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;		
	/*case TAG007://selection voxel
		dstColor = vec4(0.0, 1.0, 1.0, 0.2);
		processZvalue(geomPos, data_found);
		break;*/
		
	default: //if there wasn't a markup tag, begins depth processing 
		depth = controlTag.y;//get voxel's depth //Revised by Ting
		
		//zvalue is set once, only to the first valid voxel in ray composition.
		//so, if some valid voxel was already found, depth buffer from previous voxel
		//won't be ovewritten, ignoring this step
		if(!data_found)
		
			//1) all untagged voxels have depth=0 and must be shown.
			//2) also, voxels deeper (with greater depth) than the slicing mesh have to be shown.
			if(depth == 0. || depth_slicing <= depth)
			
				//a scalar greater than noise threshold determines a valid voxel
				if(image3Dscalar > noise_threshold)
				
					//TFscalar is what the user is seeing indeed,
					//so consider only valid entries from transfer function
					if(TFscalar > 0.0)
						processZvalue(geomPos, data_found);
		break;
	}
}

//Process the axis aligned planes, showing them and
//doing volume clipping if desired by the user
void processAxisPlanes(in bool flipPlane, in vec4 geomPos, in vec4 axialColor, 
in vec4 sagittalColor, in vec4 coronalColor, inout bool data_found, 
inout vec4 dstColor){

	//Drawing planes from side 1
	if (drawPlaneSide1[VMTK_AXIAL] == true && geomPos.y < cursorPosition.z + stepSize 
			&& geomPos.y > cursorPosition.z ) {
		dstColor += axialColor;
		processZvalue(geomPos, data_found);
	}
	if (drawPlaneSide1[VMTK_SAGITTAL] == true && geomPos.z < cursorPosition.x + stepSize
			&& geomPos.z > cursorPosition.x ) {
		dstColor += sagittalColor;
		processZvalue(geomPos, data_found);
	}
	if (drawPlaneSide1[VMTK_CORONAL] == true && geomPos.x < cursorPosition.y + stepSize
			&& geomPos.x > cursorPosition.y ) {
		dstColor += coronalColor;
		processZvalue(geomPos, data_found);
	}

	//Clipping planes from side 1
	if (!flipPlane){
		if ((clippingPlaneSide1[VMTK_AXIAL] == true && geomPos.y > cursorPosition.z  + stepSize )
			|| (clippingPlaneSide1[VMTK_SAGITTAL] == true && geomPos.z > cursorPosition.x + stepSize )
			|| (clippingPlaneSide1[VMTK_CORONAL] == true && geomPos.x > cursorPosition.y + stepSize )){

			//no color until ray is passing through clipped portion of volume
			dstColor = vec4(0.0);
			z_value = 1.0;

			//the clipped portion of volume isn't valid data to be 
			//included in ray composition, so no data found until this position
			data_found = false;
		}
	}else {
		if ((clippingPlaneSide1[VMTK_AXIAL] == true && geomPos.y < cursorPosition.z  + stepSize )
			|| (clippingPlaneSide1[VMTK_SAGITTAL] == true && geomPos.z < cursorPosition.x + stepSize )
			|| (clippingPlaneSide1[VMTK_CORONAL] == true && geomPos.x < cursorPosition.y + stepSize )){

			//no color until ray is passing through clipped portion of volume
			dstColor = vec4(0.0);
			z_value = 1.0;

			//the clipped portion of volume isn't valid data to be 
			//included in ray composition, so no data found until this position
			data_found = false;
		}
	}//EndIf FlipPlane
	
}

void main(void) {
	int i, markupTag;
	bool data_found, hit;
	float scalar, t, tnear, tfar, tag;
	vec4 geomPos, initGeomPos, volExtentMin, volExtentMax;
	vec4 axialColor, sagittalColor, coronalColor;
	vec4 dstColor, srcColor, srcColor1, srcColor2, camDir;
	vec4 geomDir, lensViewPos;
	vec4 focusPoint, lensDir;
	float depth_sample;	
	mat4 mvi;

	axialColor 		= vec4(0.0, 0.08, 0.08, 0.2);//color of axial plane
	sagittalColor 	= vec4(0.08, 0.0, 0.08, 0.2);//color of sagittal plane
	coronalColor 	= vec4(0.08, 0.08, 0.0, 0.2);//color of coronal plane

	volExtentMin = vec4(0.0, 0.0, 0.0, 0.0);//min value of boundingbox
	volExtentMax = vec4(1.0, 1.0, 1.0, 1.0);//max value of boundingbox

	dstColor = srcColor = srcColor1= srcColor2= vec4(0.0);//initialize color
	lensViewPos = vec4(1.0);	
	z_value = 1.0;	

	data_found = false; //flag used when valid voxels are found
	camDir = vec4(0.0, 0.0, -1.0, 0.0);//camera direction, used for ray composition

	mvi = gl_ModelViewMatrixInverse;//inverse model view matrix

	//determine volume entry position / ray's starting point
	initGeomPos = gl_TexCoord[1]; //world coordinate
	initGeomPos.w = 1.0;

	//compute ray direction
	geomDir = normalize(mvi * camDir) * scaleFactor;
	geomDir.w = 0.0;
	
	//lens focal point
	focusPoint.xyz 	= focalPoint.xyz;
	focusPoint.w 	= 1.0;
	focusPoint = focusPoint * scaleFactor; 		

	hit = intersectBox(initGeomPos, geomDir, volExtentMin, volExtentMax, tnear, tfar);
    if (!hit) {
       gl_FragColor = vec4(1.0,0.0,0.0,1.0);       
       return;
    }
    	
	//Recovering the depth of lens fragment based on
    //windows-position [witdh,height] and depth in [0,1]		
	lensViewPos.xy = vec2(gl_FragCoord.x,gl_FragCoord.y);	//screen coordinate	
	lensViewPos.z 	= texture2DRect(LENSDEPTHGEOMETRY, lensViewPos.xy).r; //geometry back depth texture
	lensViewPos.w 	= 1.0;
		 
    if (tnear < 0.0) tnear = 0.0; // clamp to near plane    
	t = tnear;
	
	//If is on Lens Mode and it is over volume data, display visible fragment as simple slice	
	if ( (abs(lensViewPos.z-1.0) > 0.000001)) {
								
		if (lensZooming == 0.0){	//without zooming		
			lensDir = geomDir;
		} else {
			lensDir = normalize(focusPoint - initGeomPos);
		}
						
		hit = intersectBox(initGeomPos, lensDir, volExtentMin, volExtentMax, tnear, tfar);
    	if (!hit) {
       			gl_FragColor = vec4(0.0,0.0,0.0,1.0);       			
       			return;
    	}
    		
    	if (tnear < 0.0) tnear = 0.0; // clamp to near plane    
		t = tnear;
	 		 
	 	while (t < tfar && dstColor.a < 0.99  ) {			
			
			//geomPos = volPos ;		
			geomPos = initGeomPos + lensDir * t; 
				
			scalar = texture3D(IMAGE3D, geomPos.xyz).r;//get RAW scalar value
			
			if(useSelTF){
					srcColor = texture1D(TRANSFERFUNCTIONCOLOR1D, scalar);//apply transfer function
			} else {
					srcColor = texture1D(TRANSFERFUNCTION1D, scalar);//apply transfer function
			}			
			srcColor.rgb *= srcColor.a; //compute associated color of voxel
			
			processControlTags(scalar, srcColor.a, geomPos, data_found, dstColor);
			processAxisPlanes(flipPlaneSide1, geomPos, axialColor, sagittalColor, coronalColor, data_found, dstColor);
			
			if(data_found){						
				dstColor.rgb += mix(vec3(0.0), srcColor.rgb, 1.0 - dstColor.a);			
				dstColor.a += (1.0 - dstColor.a) * srcColor.a;
				
				if (lensComposition)				
					t=tfar; //display single slice
					//break;	
								
			}//EndIf
					
			//advance ray position along ray direction
			t += stepSize;					
		}//EndWhile
						
	} 	else {
	
		// ray traversal loop 
		while (t < tfar && dstColor.a < 0.99  ) {
		
			geomPos = initGeomPos + geomDir * t;
			
			scalar = texture3D(IMAGE3D, geomPos.xyz).r;//get RAW scalar value
			srcColor = texture1D(TRANSFERFUNCTION1D, scalar);//apply transfer function
			srcColor.rgb *= srcColor.a; //compute associated color of voxel

			//control volume processing determines depth buffer and
			//the beginning of ray composition
			processControlTags(scalar, srcColor.a, geomPos, data_found, dstColor);
			processAxisPlanes(flipPlaneSide1, geomPos, axialColor, sagittalColor, coronalColor, data_found, dstColor);
			
			// If a valid voxel was found, begin/continue front-to-back ray composition
			if(data_found){				
				//color composition
				dstColor.rgb += mix(vec3(0.0), srcColor.rgb, 1.0 - dstColor.a);			
				dstColor.a += (1.0 - dstColor.a) * srcColor.a;				
			}
					
			//advance ray position along ray direction
			t += stepSize;
		}//End while
		 
	}//End If
	
	if (dstColor.a == 0.0) 				
	  gl_FragColor = vec4(0.5,0.5,0.5,1.0);
	else
	  gl_FragColor = dstColor;
			
}
